@@ -1,15 +1,14 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "ellama" "1.12.18"
+(define-package "ellama" "20250918.1606"
   "Tool for interacting with LLMs."
   '((emacs     "28.1")
     (llm       "0.24.0")
     (plz       "0.8")
     (transient "0.7")
-    (compat    "29.1")
-    (yaml      "1.2.3"))
-  :url "https://github.com/s-kostyaev/ellama"
-  :commit "c9cf0ae36cac6d4a65bafb8f947e3195e2e2f5fd"
-  :revdesc "c9cf0ae36cac"
+    (compat    "29.1"))
+  :url "http://github.com/s-kostyaev/ellama"
+  :commit "2857f85b8f10eb587afb8cb3da1d96cb236b2856"
+  :revdesc "2857f85b8f10"
   :keywords '("help" "local" "tools")
   :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com"))
   :maintainers '(("Sergey Kostyaev" . "sskostyaev@gmail.com")))
