@@ -1,12 +1,13 @@
 ;;; ellama.el --- Tool for interacting with LLMs -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023-2026  Free Software Foundation, Inc.
+;; Copyright (C) 2023-2025  Free Software Foundation, Inc.
 
 ;; Author: Sergey Kostyaev <sskostyaev@gmail.com>
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
-;; Package-Requires: ((emacs "28.1") (llm "0.24.0") (plz "0.8") (transient "0.7") (compat "29.1") (yaml "1.2.3"))
-;; Version: 1.12.15
+;; Package-Requires: ((emacs "28.1") (llm "0.24.0") (plz "0.8") (transient "0.7") (compat "29.1"))
+;; Package-Version: 20250918.1606
+;; Package-Revision: 2857f85b8f10
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -40,8 +41,6 @@
 (require 'llm-provider-utils)
 (require 'compat)
 (eval-when-compile (require 'rx))
-(require 'ellama-tools)
-(require 'ellama-skills)
 
 (defgroup ellama nil
   "Tool for interacting with LLMs."
@@ -227,15 +226,6 @@ PROMPT is a prompt string."
 	  (const :tag "By generating name with reasoning LLM based on prompt." ellama-generate-name-by-reasoning-llm)
           (function :tag "By custom function")))
 
-(defcustom ellama-response-process-method 'streaming
-  "How to process LLM response.
-If the default streaming method is too resource-heavy, you can try other
-options."
-  :type `(choice
-	  (const :tag "Streaming" streaming)
-	  (const :tag "Async" async)
-	  (integer :tag "Skip every N messages before process one")))
-
 (defcustom ellama-define-word-prompt-template "Define %s"
   "Prompt template for `ellama-define-word'."
   :type 'string)
@@ -468,15 +458,6 @@ It should be a function with single argument generated text string."
   "Enable debug."
   :type 'boolean)
 
-(defcustom ellama-sessions-directory (file-truename
-				      (file-name-concat
-				       user-emacs-directory
-				       "ellama-sessions"))
-  "Directory for saved ellama sessions."
-  :type 'string)
-
-(defvar ellama--current-session-id nil)
-
 (defun ellama--set-file-name-and-save ()
   "Set buffer file name and save buffer."
   (interactive)
@@ -539,56 +520,24 @@ It should be a function with single argument generated text string."
 (defun ellama--replace-first-begin-src (text)
   "Replace first begin src in TEXT."
   (if (not (string-match-p (rx (literal "#+BEGIN_SRC")) text))
-      (with-temp-buffer
-        (insert text)
-        (goto-char (point-min))
-        (when (re-search-forward "^[[:space:]]*```" nil t)
-          (replace-match "#+BEGIN_SRC" t t))
-        (buffer-substring-no-properties (point-min) (point-max)))
+      (replace-regexp-in-string "^[[:space:]]*```\\(\\(.\\|\n\\)*\\)" "#+BEGIN_SRC\\1" text)
     text))
 
-(defun ellama--replace-inline-code-fences (text)
-  "Replace inline markdown code fences in TEXT with org equivalents."
-  (with-temp-buffer
-    (insert text)
-    (goto-char (point-min))
-    (while (not (eobp))
-      (let* ((line-beg (line-beginning-position))
-             (line-end (line-end-position))
-             (line (buffer-substring-no-properties line-beg line-end))
-             (fence-pos (string-match "```" line)))
-        ;; Handle cases like `text ```lang' and `text ```text'.
-        (when (and fence-pos (> fence-pos 0))
-          (let ((prefix (substring line 0 fence-pos))
-                (suffix (substring line (+ fence-pos 3))))
-            (cond
-             ((string-match-p "\\`[A-Za-z0-9-]+\\'" suffix)
-              (goto-char line-beg)
-              (delete-region line-beg line-end)
-              (insert prefix "\n#+BEGIN_SRC " suffix))
-             ((not (string= suffix ""))
-              (goto-char line-beg)
-              (delete-region line-beg line-end)
-              (insert prefix "\n#+END_SRC\n" suffix))))))
-      (forward-line 1))
-    (buffer-substring-no-properties (point-min) (point-max))))
-
 (defun ellama--replace-bad-code-blocks (text)
   "Replace code src blocks in TEXT."
   (with-temp-buffer
     (insert (propertize text 'hard t))
-    (let ((open-blocks 0)
-          (pattern
-           "^\\([[:blank:]]*\\)#\\+\\(BEGIN_SRC\\|END_SRC\\)\\(?:[[:blank:]].*\\)?$"))
-      (goto-char (point-min))
-      (while (re-search-forward pattern nil t)
-        (if (string= (match-string 2) "BEGIN_SRC")
-            (setq open-blocks (1+ open-blocks))
-          (if (> open-blocks 0)
-              (setq open-blocks (1- open-blocks))
-            (let ((indent (match-string 1)))
-              (replace-match (concat indent "#+BEGIN_SRC") t t))
-            (setq open-blocks (1+ open-blocks))))))
+    (goto-char (point-min))
+    ;; skip good code blocks
+    (while (re-search-forward "#\\+BEGIN_SRC\\(.\\|\n\\)*?#\\+END_SRC" nil t))
+    (while (re-search-forward "#\\+END_SRC\\(\\(.\\|\n\\)*?\\)#\\+END_SRC" nil t)
+      (unless (string-match-p "#\\+BEGIN_SRC" (match-string 1))
+	(replace-match "#+BEGIN_SRC\\1#+END_SRC")))
+    (goto-char (match-beginning 0))
+    (while (re-search-backward "#\\+END_SRC\\(\\(.\\|\n\\)*?\\)#\\+END_SRC" nil t)
+      (unless (string-match-p "#\\+BEGIN_SRC" (match-string 1))
+	(replace-match "#+BEGIN_SRC\\1#+END_SRC"))
+      (goto-char (match-beginning 0)))
     (buffer-substring-no-properties (point-min) (point-max))))
 
 (defun ellama--replace (from to beg end)
@@ -696,7 +645,8 @@ This filter contains only subset of markdown syntax to be good enough."
     text
     ;; code blocks
     (replace-regexp-in-string "^[[:space:]]*```\\(.+\\)$" "#+BEGIN_SRC \\1")
-    (ellama--replace-inline-code-fences)
+    (replace-regexp-in-string "^\\(.+\\)```\\([A-Za-z0-9\\-]+\\)$" "\\1\n#+BEGIN_SRC \\2")
+    (replace-regexp-in-string "^\\(.+\\)```\\(.+\\)$" "\\1\n#+END_SRC\n\\2")
     (ellama--replace-first-begin-src)
     (replace-regexp-in-string "^<!-- language: \\(.+\\) -->\n```" "#+BEGIN_SRC \\1")
     (replace-regexp-in-string "^[[:space:]]*```$" "#+END_SRC")
@@ -717,6 +667,13 @@ This filter contains only subset of markdown syntax to be good enough."
 	   ;; If ellama-enable-keymap is nil, remove the key bindings
 	   (define-key global-map (kbd ellama-keymap-prefix) nil))))
 
+(defcustom ellama-sessions-directory (file-truename
+				      (file-name-concat
+				       user-emacs-directory
+				       "ellama-sessions"))
+  "Directory for saved ellama sessions."
+  :type 'string)
+
 (defcustom ellama-naming-provider nil
   "LLM provider for generating names."
   :type '(sexp :validate llm-standard-provider-p))
@@ -727,6 +684,8 @@ This filter contains only subset of markdown syntax to be good enough."
 
 (defvar-local ellama--current-session nil)
 
+(defvar ellama--current-session-id nil)
+
 (defcustom ellama-session-line-template " ellama session: %s"
   "Template for formatting the current session line."
   :type 'string)
@@ -739,6 +698,14 @@ This filter contains only subset of markdown syntax to be good enough."
 			ellama--current-session-id))
 	      'face 'ellama-face))
 
+(defun ellama-session-update-header-line ()
+  "Update header line for ellama session header line mode."
+  (when (listp header-line-format)
+    (let ((element '(:eval (ellama-session-line))))
+      (if ellama-session-header-line-mode
+          (add-to-list 'header-line-format element t)
+	(setq header-line-format (delete element header-line-format))))))
+
 ;;;###autoload
 (define-minor-mode ellama-session-header-line-mode
   "Toggle Ellama Session header line mode."
@@ -750,6 +717,14 @@ This filter contains only subset of markdown syntax to be good enough."
   ellama-session-header-line-mode
   ellama-session-header-line-mode)
 
+(defun ellama-session-update-mode-line ()
+  "Update mode line for ellama session mode line mode."
+  (when (listp mode-line-format)
+    (let ((element '(:eval (ellama-session-line))))
+      (if ellama-session-mode-line-mode
+	  (add-to-list 'mode-line-format element t)
+	(setq mode-line-format (delete element mode-line-format))))))
+
 ;;;###autoload
 (define-minor-mode ellama-session-mode-line-mode
   "Toggle Ellama Session mode line mode."
@@ -761,22 +736,6 @@ This filter contains only subset of markdown syntax to be good enough."
   ellama-session-mode-line-mode
   ellama-session-mode-line-mode)
 
-(defun ellama-session-update-header-line ()
-  "Update header line for ellama session header line mode."
-  (when (listp header-line-format)
-    (let ((element '(:eval (ellama-session-line))))
-      (if ellama-session-header-line-mode
-          (add-to-list 'header-line-format element t)
-	(setq header-line-format (delete element header-line-format))))))
-
-(defun ellama-session-update-mode-line ()
-  "Update mode line for ellama session mode line mode."
-  (when (listp mode-line-format)
-    (let ((element '(:eval (ellama-session-line))))
-      (if ellama-session-mode-line-mode
-	  (add-to-list 'mode-line-format element t)
-	(setq mode-line-format (delete element mode-line-format))))))
-
 (defvar ellama--active-sessions (make-hash-table :test #'equal))
 
 (cl-defstruct ellama-session
@@ -892,8 +851,7 @@ Defaults to md, but supports org.  Depends on `ellama-major-mode'."
   "Create new ellama session with unique id.
 Provided PROVIDER and PROMPT will be used in new session.
 If EPHEMERAL non nil new session will not be associated with any file."
-  (let* ((dir default-directory)
-	 (name (ellama-generate-name provider 'ellama prompt))
+  (let* ((name (ellama-generate-name provider 'ellama prompt))
 	 (count 1)
 	 (name-with-suffix (format "%s %d" name count))
 	 (id (if (and (not (ellama-get-session-buffer name))
@@ -914,7 +872,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
 		       ellama-sessions-directory
 		       (concat id "." (ellama-get-session-file-extension)))))
 	 (session (make-ellama-session
-		   :id id :provider provider :file file-name :extra `(:dir ,dir)))
+		   :id id :provider provider :file file-name))
 	 (buffer (if file-name
 		     (progn
 		       (make-directory ellama-sessions-directory t)
@@ -923,7 +881,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (setq ellama--current-session-id id)
     (puthash id buffer ellama--active-sessions)
     (with-current-buffer buffer
-      (setq default-directory dir)
+      (setq default-directory ellama-sessions-directory)
       (funcall ellama-major-mode)
       (setq ellama--current-session session)
       (ellama-session-mode +1))
@@ -1053,13 +1011,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	       :provider (ellama-session-provider session)
 	       :file (ellama-session-file session)
 	       :prompt (ellama-session-prompt session)
-	       :extra extra))
-	(with-current-buffer buffer
-	  (when-let* ((extra)
-		      ((plistp extra))
-		      (dir (plist-get extra :dir))
-		      ((file-exists-p dir)))
-	    (setq default-directory dir))))
+	       :extra extra)))
       (setq ellama--current-session-id (ellama-session-id ellama--current-session))
       (puthash (ellama-session-id ellama--current-session)
 	       buffer ellama--active-sessions)
@@ -1244,45 +1196,6 @@ Otherwire return current active session."
 
 (defvar ellama-global-system nil)
 
-(defun ellama-get-agents-md-path ()
-  "Search for AGENTS.md file from current directory up to project root.
-Returns the full path to AGENTS.md if found, or nil if not found."
-  (let* ((current-dir (file-name-directory (expand-file-name default-directory)))
-         (project-root (or (and (ellama-tools-project-root-tool)
-				(file-name-directory (expand-file-name
-						      (ellama-tools-project-root-tool))))
-			   current-dir))
-         found-path)
-    ;; Walk up from current directory to project root
-    (while (and (not found-path)
-		(or (string= current-dir project-root)
-		    (string> current-dir project-root)))
-      (let ((agents-path (expand-file-name "AGENTS.md" current-dir)))
-        (when (file-exists-p agents-path)
-          (setq found-path agents-path)))
-      ;; Move to parent directory
-      (setq current-dir (file-name-directory (expand-file-name ".." current-dir))))
-    found-path))
-
-(defun ellama-get-agents-md ()
-  "Return current project or subproject AGENTS.md content."
-  (or (when-let ((path (ellama-get-agents-md-path)))
-	(concat
-	 "\n"
-	 (with-temp-buffer
-	   (insert-file-contents path)
-	   (buffer-string))))
-      ""))
-
-(defun ellama-get-system-message (&optional msg)
-  "Return the effective system message, including dynamically scanned skills.
-MSG is a text that will be included in the resulting system message."
-  (let ((msg (concat (or msg ellama-global-system "")
-		     (ellama-get-agents-md)
-		     (ellama-skills-generate-prompt))))
-    (when (not (string= msg ""))
-      msg)))
-
 (defvar-local ellama--stop-scroll nil)
 
 ;;;###autoload
@@ -1417,8 +1330,7 @@ INSERT-REASONING is a function for reasoning insertion.
 REASONING-BUFFER is a buffer for reasoning."
   (lambda (response)
     (let ((text (plist-get response :text))
-	  (reasoning (plist-get response :reasoning))
-	  (tool-results (plist-get response :tool-results)))
+	  (reasoning (plist-get response :reasoning)))
       (funcall
        insert-text
        (concat
@@ -1426,7 +1338,7 @@ REASONING-BUFFER is a buffer for reasoning."
 	  (if
 	      (or (not ellama-output-remove-reasoning)
 		  ellama--current-session)
-	      (concat "<think>\n" reasoning "\n</think>\n")
+	      (concat "<think>\n" reasoning)
 	    (progn
 	      (with-current-buffer reasoning-buffer
 		(funcall insert-reasoning reasoning)
@@ -1436,134 +1348,10 @@ REASONING-BUFFER is a buffer for reasoning."
 		   (when ellama-reasoning-display-action-function
 		     `((ignore . (,ellama-reasoning-display-action-function)))))))
 	      nil)))
-	(when tool-results
-	  (format "\n<think>\n%s\n</think>\n"
-		  tool-results))
 	(when text
-	  (string-trim text)))))))
-
-(defun ellama--tool-call-error-p (err-type)
-  "Return non-nil when ERR-TYPE indicates a tool call error."
-  (and err-type
-       (memq 'llm-tool-call-error
-	     (get err-type 'error-conditions))))
-
-(defun ellama--append-tool-error-to-prompt (prompt msg)
-  "Append tool call error MSG to PROMPT."
-  (when prompt
-    (llm-chat-prompt-append-response
-     prompt
-     (if (stringp msg)
-	 msg
-       (format "%s" (or msg "Unknown tool call error")))
-     'system)))
-
-(defun ellama--error-handler (buffer errcb &optional prompt
-				     retry-fn)
-  "Error handler function.
-BUFFER is the current ellama buffer.
-ERRCB is an error callback.
-PROMPT is the active prompt.
-RETRY-FN is called to retry the request."
-  (lambda (err-type msg)
-    (with-current-buffer buffer
-      (if (and retry-fn
-	       prompt
-	       (ellama--tool-call-error-p err-type))
-	  (progn
-	    (ellama--append-tool-error-to-prompt prompt msg)
-	    (funcall retry-fn))
-	(cancel-change-group ellama--change-group)
-	(when ellama-spinner-enabled
-	  (spinner-stop))
-	(funcall errcb msg)
-	(setq ellama--current-request nil)
-	(ellama-request-mode -1)))))
-
-(defun ellama--response-handler (result-handler reasoning-buffer buffer donecb errcb provider llm-prompt async filter)
-  "Response handler function.
-RESULT-HANDLER handles text insertion.
-REASONING-BUFFER used for reasoning output.
-BUFFER is the current ellama buffer.
-DONECB is a done callback.
-PROVIDER is an llm provider.
-ERRCB is an error callback.
-LLM-PROMPT is current llm prompt.
-ASYNC flag is for asyncronous requests.
-FILTER is a function that's applied to (partial) response strings before they're
-inserted into the BUFFER."
-  (lambda (response)
-    (let ((text (plist-get response :text))
-	  (reasoning (plist-get response :reasoning))
-	  (tool-result (plist-get response :tool-results)))
-      (funcall result-handler response)
-      (when (or ellama--current-session
-		(not reasoning))
-	(when (not tool-result) (kill-buffer reasoning-buffer)))
-      (if tool-result
-	  (cl-labels
-	      ((start-request ()
-		 (let* ((insert-text
-			 (ellama--insert
-			  buffer
-			  (with-current-buffer buffer
-			    (if ellama--current-session
-				(point-max)
-			      (point)))
-			  filter))
-			(insert-reasoning
-			 (ellama--insert reasoning-buffer nil
-					 #'ellama--translate-markdown-to-org-filter))
-			(handler
-			 (ellama--handle-partial
-			  insert-text insert-reasoning reasoning-buffer))
-			(cnt 0)
-			(skip-handler
-			 (lambda (request)
-			   (if (= cnt ellama-response-process-method)
-			       (progn
-				 (funcall handler request)
-				 (setq cnt 0))
-			     (cl-incf cnt))))
-			(error-handler
-			 (ellama--error-handler
-			  buffer errcb llm-prompt
-			  (lambda ()
-			    (start-request)))))
-		   (with-current-buffer buffer
-		     (if async
-			 (llm-chat-async
-			  provider
-			  llm-prompt
-			  (ellama--response-handler
-			   handler reasoning-buffer buffer donecb errcb provider
-			   llm-prompt async filter)
-			  error-handler
-			  t)
-		       (llm-chat-streaming
-			provider
-			llm-prompt
-			(if (integerp ellama-response-process-method)
-			    skip-handler handler)
-			(ellama--response-handler
-			 handler reasoning-buffer buffer donecb errcb provider
-			 llm-prompt async filter)
-			error-handler
-			t))))))
-	    (start-request))
-	(with-current-buffer buffer
-	  (accept-change-group ellama--change-group)
-	  (when ellama-spinner-enabled
-	    (spinner-stop))
-	  (if (and (listp donecb)
-		   (functionp (car donecb)))
-	      (mapc (lambda (fn) (funcall fn text))
-		    donecb)
-	    (funcall donecb text))
-	  (when ellama-session-hide-org-quotes
-	    (ellama-collapse-org-quotes))
-	  (setq ellama--current-request nil)
-	  (ellama-request-mode -1))))))
+	  (if (and reasoning ellama--current-session)
+	      (concat "</think>\n" (string-trim text))
+	    (string-trim text))))))))
 
 (defun ellama-stream (prompt &rest args)
   "Query ellama for PROMPT.
@@ -1576,8 +1364,6 @@ in.  Default value is (current-buffer).
 
 :point POINT -- POINT is the point in buffer to insert ellama reply at.
 
-:tools LIST -- LIST of enabled in the current session tools.
-
 :filter FILTER -- FILTER is a function that's applied to (partial) response
 strings before they're inserted into the BUFFER.
 
@@ -1622,91 +1408,71 @@ failure (with BUFFER current).
 		      (error "Error calling the LLM: %s" msg))))
 	 (donecb (or (plist-get args :on-done) #'ignore))
 	 (prompt-with-ctx (ellama-context-prompt-with-context prompt))
-	 (system (ellama-get-system-message (plist-get args :system)))
-	 (session-tools (and session
-                             (ellama-session-extra session)
-                             (plist-get (ellama-session-extra session) :tools)))
-	 (tools (or session-tools
-		    (plist-get args :tools)
-		    ellama-tools-enabled))
+	 (system (or (plist-get args :system)
+		     ellama-global-system))
 	 (llm-prompt (if session
 			 (if (llm-chat-prompt-p (ellama-session-prompt session))
 			     (progn
 			       (llm-chat-prompt-append-response
 				(ellama-session-prompt session)
 				prompt-with-ctx)
-			       (setf (llm-chat-prompt-tools (ellama-session-prompt session))
-				     tools)
-			       ;; System message is part of prompt context and should not be
-			       ;; appended on each interaction.
+			       (when system
+				 (llm-chat-prompt-append-response
+				  (ellama-session-prompt session)
+				  system 'system))
 			       (ellama-session-prompt session))
 			   (setf (ellama-session-prompt session)
-				 (llm-make-chat-prompt prompt-with-ctx :context system
-						       :tools tools)))
-		       (llm-make-chat-prompt prompt-with-ctx :context system
-					     :tools tools))))
+				 (llm-make-chat-prompt prompt-with-ctx :context system)))
+		       (llm-make-chat-prompt prompt-with-ctx :context system))))
     (with-current-buffer reasoning-buffer
       (org-mode))
     (with-current-buffer buffer
       (ellama-request-mode +1)
-      (cl-labels
-	  ((start-request ()
-	     (let* ((insert-text
-		     (ellama--insert buffer point filter))
-		    (insert-reasoning
-		     (ellama--insert reasoning-buffer nil
-				     #'ellama--translate-markdown-to-org-filter))
-		    (handler
-		     (ellama--handle-partial
-		      insert-text insert-reasoning reasoning-buffer))
-		    (error-handler
-		     (ellama--error-handler
-		      buffer errcb llm-prompt
-		      #'start-request))
-		    (request (pcase ellama-response-process-method
-			       ('async (llm-chat-async
-					provider
-					llm-prompt
-					(ellama--response-handler
-					 handler reasoning-buffer buffer donecb errcb provider
-					 llm-prompt t filter)
-					error-handler
-					t))
-			       ('streaming (llm-chat-streaming
-					    provider
-					    llm-prompt
-					    handler
-					    (ellama--response-handler
-					     handler reasoning-buffer buffer donecb errcb
-					     provider llm-prompt nil filter)
-					    error-handler
-					    t))
-			       ((pred integerp)
-				(let* ((cnt 0)
-				       (skip-handler
-					(lambda (request)
-					  (if (= cnt ellama-response-process-method)
-					      (progn
-						(funcall handler request)
-						(setq cnt 0))
-					    (cl-incf cnt)))))
-				  (llm-chat-streaming
-				   provider
-				   llm-prompt
-				   skip-handler
-				   (ellama--response-handler
-				    handler reasoning-buffer buffer donecb errcb provider
-				    llm-prompt t filter)
-				   error-handler
-				   t))))))
-	       (setq ellama--change-group (prepare-change-group))
-	       (activate-change-group ellama--change-group)
-	       (when ellama-spinner-enabled
-		 (require 'spinner)
-		 (spinner-start ellama-spinner-type))
-	       (with-current-buffer buffer
-		 (setq ellama--current-request request)))))
-	(start-request)))))
+      (let* ((insert-text
+	      (ellama--insert buffer point filter))
+	     (insert-reasoning
+	      (ellama--insert reasoning-buffer nil #'ellama--translate-markdown-to-org-filter)))
+	(setq ellama--change-group (prepare-change-group))
+	(activate-change-group ellama--change-group)
+	(when ellama-spinner-enabled
+	  (require 'spinner)
+	  (spinner-start ellama-spinner-type))
+	(let* ((handler (ellama--handle-partial insert-text insert-reasoning reasoning-buffer))
+	       (request (llm-chat-streaming
+			 provider
+			 llm-prompt
+			 handler
+			 (lambda (response)
+			   (let ((text (plist-get response :text))
+				 (reasoning (plist-get response :reasoning)))
+			     (funcall handler response)
+			     (when (or ellama--current-session
+				       (not reasoning))
+			       (kill-buffer reasoning-buffer))
+			     (with-current-buffer buffer
+			       (accept-change-group ellama--change-group)
+			       (when ellama-spinner-enabled
+				 (spinner-stop))
+			       (if (and (listp donecb)
+					(functionp (car donecb)))
+				   (mapc (lambda (fn) (funcall fn text))
+					 donecb)
+				 (funcall donecb text))
+			       (when ellama-session-hide-org-quotes
+				 (ellama-collapse-org-quotes))
+			       (setq ellama--current-request nil)
+			       (ellama-request-mode -1))))
+			 (lambda (_ msg)
+			   (with-current-buffer buffer
+			     (cancel-change-group ellama--change-group)
+			     (when ellama-spinner-enabled
+			       (spinner-stop))
+			     (funcall errcb msg)
+			     (setq ellama--current-request nil)
+			     (ellama-request-mode -1)))
+			 t)))
+	  (with-current-buffer buffer
+	    (setq ellama--current-request request)))))))
 
 (defun ellama-chain (initial-prompt forms &optional acc)
   "Call chain of FORMS on INITIAL-PROMPT.
